import curve from './curve.js';
import * as util from './util.js';

var crypto = window.crypto;

if (!crypto || !crypto.subtle || typeof crypto.getRandomValues !== 'function') {
  throw new Error('WebCrypto not found');
}

export const cryptoInternals = {
  getRandomBytes: function (size) {
    var array = new Uint8Array(size);
    crypto.getRandomValues(array);
    return array.buffer;
  },
  encrypt: function (key, data, iv) {
    return crypto.subtle
      .importKey('raw', key, { name: 'AES-CBC' }, false, ['encrypt'])
      .then(function (key) {
        return crypto.subtle.encrypt({ name: 'AES-CBC', iv: new Uint8Array(iv) }, key, data);
      });
  },
  decrypt: function (key, data, iv) {
    return crypto.subtle
      .importKey('raw', key, { name: 'AES-CBC' }, false, ['decrypt'])
      .then(function (key) {
        return crypto.subtle.decrypt({ name: 'AES-CBC', iv: new Uint8Array(iv) }, key, data);
      });
  },
  sign: function (key, data) {
    return crypto.subtle
      .importKey('raw', key, { name: 'HMAC', hash: { name: 'SHA-256' } }, false, ['sign'])
      .then(function (key) {
        return crypto.subtle.sign({ name: 'HMAC', hash: 'SHA-256' }, key, data);
      });
  },

  hash: function (data) {
    return crypto.subtle.digest({ name: 'SHA-512' }, data);
  },

  HKDF: function (input, salt, info) {
    // Specific implementation of RFC 5869 that only returns the first 3 32-byte chunks
    // TODO: We dont always need the third chunk, we might skip it
    return cryptoInternals.sign(salt, input).then(function (PRK) {
      var infoBuffer = new ArrayBuffer(info.byteLength + 1 + 32);
      var infoArray = new Uint8Array(infoBuffer);
      infoArray.set(new Uint8Array(info), 32);
      infoArray[infoArray.length - 1] = 1;
      return cryptoInternals.sign(PRK, infoBuffer.slice(32)).then(function (T1) {
        infoArray.set(new Uint8Array(T1));
        infoArray[infoArray.length - 1] = 2;
        return cryptoInternals.sign(PRK, infoBuffer).then(function (T2) {
          infoArray.set(new Uint8Array(T2));
          infoArray[infoArray.length - 1] = 3;
          return cryptoInternals.sign(PRK, infoBuffer).then(function (T3) {
            return [T1, T2, T3];
          });
        });
      });
    });
  },

  // Curve 25519 crypto
  createKeyPair: function (privKey) {
    if (privKey === undefined) {
      privKey = cryptoInternals.getRandomBytes(32);
    }
    return curve.createKeyPair(privKey);
  },
  ECDHE: function (pubKey, privKey) {
    return curve.ECDHE(pubKey, privKey);
  },
  Ed25519Sign: function (privKey, message) {
    return curve.Ed25519Sign(privKey, message);
  },
  Ed25519Verify: function (pubKey, msg, sig) {
    return curve.Ed25519Verify(pubKey, msg, sig);
  },
};

// HKDF for TextSecure has a bit of additional handling - salts always end up being 32 bytes
export const _HKDF = function (input, salt, info) {
  if (salt.byteLength != 32) {
    throw new Error('Got salt of incorrect length');
  }

  return cryptoInternals.HKDF(input, salt, util.toArrayBuffer(info));
};

export const verifyMAC = function (data, key, mac, length) {
  return cryptoInternals.sign(key, data).then(function (calculated_mac) {
    if (mac.byteLength != length || calculated_mac.byteLength < length) {
      throw new Error('Bad MAC length');
    }
    var a = new Uint8Array(calculated_mac);
    var b = new Uint8Array(mac);
    var result = 0;
    for (var i = 0; i < mac.byteLength; ++i) {
      result = result | (a[i] ^ b[i]);
    }
    if (result !== 0) {
      throw new Error('Bad MAC');
    }
  });
};

export const HKDF = {
  deriveSecrets: function (input, salt, info) {
    return _HKDF(input, salt, info);
  },
};

export default {
  encrypt: function (key, data, iv) {
    return cryptoInternals.encrypt(key, data, iv);
  },
  decrypt: function (key, data, iv) {
    return cryptoInternals.decrypt(key, data, iv);
  },
  calculateMAC: function (key, data) {
    return cryptoInternals.sign(key, data);
  },
  verifyMAC: function (data, key, mac, length) {
    return verifyMAC(data, key, mac, length);
  },
  getRandomBytes: function (size) {
    return cryptoInternals.getRandomBytes(size);
  },
};
