import curve25519 from './curve25519_wrapper.js';

function validatePrivKey(privKey) {
  if (privKey === undefined || !(privKey instanceof ArrayBuffer) || privKey.byteLength != 32) {
    throw new Error('Invalid private key');
  }
}
function validatePubKeyFormat(pubKey) {
  if (
    pubKey === undefined ||
    ((pubKey.byteLength != 33 || new Uint8Array(pubKey)[0] != 5) && pubKey.byteLength != 32)
  ) {
    throw new Error('Invalid public key');
  }
  if (pubKey.byteLength == 33) {
    return pubKey.slice(1);
  } else {
    console.error(
      'WARNING: Expected pubkey of length 33, please report the ST and client that generated the pubkey'
    );
    return pubKey;
  }
}

function processKeys(raw_keys) {
  // prepend version byte
  var origPub = new Uint8Array(raw_keys.pubKey);
  var pub = new Uint8Array(33);
  pub.set(origPub, 1);
  pub[0] = 5;

  return { pubKey: pub.buffer, privKey: raw_keys.privKey };
}

class Curve {
  createKeyPairSync(privKey) {
    validatePrivKey(privKey);
    var raw_keys = curve25519.keyPair(privKey);
    return processKeys(raw_keys);
  }

  createKeyPair(privKey) {
    validatePrivKey(privKey);

    const keyPromise = this.worker ? this.worker.keyPair(privKey) : curve25519.keyPair(privKey);
    return keyPromise.then(processKeys);
  }

  ECDHESync(pubKey, privKey) {
    pubKey = validatePubKeyFormat(pubKey);
    validatePrivKey(privKey);

    if (pubKey === undefined || pubKey.byteLength != 32) {
      throw new Error('Invalid public key');
    }

    return curve25519.sharedSecretSync(pubKey, privKey);
  }
  ECDHE(pubKey, privKey) {
    pubKey = validatePubKeyFormat(pubKey);
    validatePrivKey(privKey);

    if (pubKey === undefined || pubKey.byteLength != 32) {
      throw new Error('Invalid public key');
    }

    if (this.worker) return this.worker.sharedSecret(pubKey, privKey);
    return curve25519.sharedSecret(pubKey, privKey);
  }

  Ed25519SignSync(privKey, message) {
    validatePrivKey(privKey);

    if (message === undefined) {
      throw new Error('Invalid message');
    }

    return curve25519.signSync(privKey, message);
  }

  Ed25519Sign(privKey, message) {
    validatePrivKey(privKey);

    if (message === undefined) {
      throw new Error('Invalid message');
    }

    if (this.worker) return this.worker.sign(privKey, message);
    return curve25519.sign(privKey, message);
  }

  Ed25519VerifySync(pubKey, msg, sig) {
    pubKey = validatePubKeyFormat(pubKey);

    if (pubKey === undefined || pubKey.byteLength != 32) {
      throw new Error('Invalid public key');
    }

    if (msg === undefined) {
      throw new Error('Invalid message');
    }

    if (sig === undefined || sig.byteLength != 64) {
      throw new Error('Invalid signature');
    }

    return curve25519.verifySync(pubKey, msg, sig);
  }

  Ed25519Verify(pubKey, msg, sig) {
    pubKey = validatePubKeyFormat(pubKey);

    if (pubKey === undefined || pubKey.byteLength != 32) {
      throw new Error('Invalid public key');
    }

    if (msg === undefined) {
      throw new Error('Invalid message');
    }

    if (sig === undefined || sig.byteLength != 64) {
      throw new Error('Invalid signature');
    }

    if (this.worker) return this.worker.verify(pubKey, msg, sig);
    return curve25519.verify(pubKey, msg, sig);
  }

  // TODO: write tests for these or get rid of them if we don't need them
  //     generateKeyPair: function () {
  //       var privKey = cryptoInternals.getRandomBytes(32);
  //       return curve.createKeyPair(privKey);
  //     },
  //     createKeyPair: function (privKey) {
  //       return curve.createKeyPair(privKey);
  //     },
  //     calculateAgreement: function (pubKey, privKey) {
  //       return curve.ECDHE(pubKey, privKey);
  //     },
  //     verifySignature: function (pubKey, msg, sig) {
  //       return curve.Ed25519Verify(pubKey, msg, sig);
  //     },
  //     calculateSignature: function (privKey, message) {
  //       return curve.Ed25519Sign(privKey, message);
}

export default new Curve();
