import { Module } from 'exports-loader?exports=Module!./curve25519';

// Insert some bytes into the emscripten memory and return a pointer
function _allocate(bytes) {
  var address = Module._malloc(bytes.length);
  Module.HEAPU8.set(bytes, address);

  return address;
}

function _readBytes(address, length, array) {
  array.set(Module.HEAPU8.subarray(address, address + length));
}

var basepoint = new Uint8Array(32);
basepoint[0] = 9;

// this promise resolves when the wasm module is initialized
const initPromise = new Promise(resolve => (Module['onRuntimeInitialized'] = resolve));

class Curve25519Wrapper {
  keyPair(privKey) {
    return initPromise.then(
      () =>
        new Promise(resolve => {
          resolve(this.keyPairSync(privKey));
        })
    );
  }
  keyPairSync(privKey) {
    var priv = new Uint8Array(privKey);
    priv[0] &= 248;
    priv[31] &= 127;
    priv[31] |= 64;

    // Where to store the result
    var publicKey_ptr = Module._malloc(32);

    // Get a pointer to the private key
    var privateKey_ptr = _allocate(priv);

    // The basepoint for generating public keys
    var basepoint_ptr = _allocate(basepoint);

    // The return value is just 0, the operation is done in place
    Module._curve25519_donna(publicKey_ptr, privateKey_ptr, basepoint_ptr);

    var res = new Uint8Array(32);
    _readBytes(publicKey_ptr, 32, res);

    Module._free(publicKey_ptr);
    Module._free(privateKey_ptr);
    Module._free(basepoint_ptr);

    return { pubKey: res.buffer, privKey: priv.buffer };
  }

  sharedSecretSync(pubKey, privKey) {
    // Where to store the result
    var sharedKey_ptr = Module._malloc(32);

    // Get a pointer to our private key
    var privateKey_ptr = _allocate(new Uint8Array(privKey));

    // Get a pointer to their public key, the basepoint when you're
    // generating a shared secret
    var basepoint_ptr = _allocate(new Uint8Array(pubKey));

    // Return value is 0 here too of course
    Module._curve25519_donna(sharedKey_ptr, privateKey_ptr, basepoint_ptr);

    var res = new Uint8Array(32);
    _readBytes(sharedKey_ptr, 32, res);

    Module._free(sharedKey_ptr);
    Module._free(privateKey_ptr);
    Module._free(basepoint_ptr);

    return res.buffer;
  }
  sharedSecret(pubKey, privKey) {
    return initPromise.then(
      () =>
        new Promise(resolve => {
          resolve(this.sharedSecretSync(pubKey, privKey));
        })
    );
  }

  signSync(privKey, message) {
    // Where to store the result
    var signature_ptr = Module._malloc(64);

    // Get a pointer to our private key
    var privateKey_ptr = _allocate(new Uint8Array(privKey));

    // Get a pointer to the message
    var message_ptr = _allocate(new Uint8Array(message));

    Module._curve25519_sign(signature_ptr, privateKey_ptr, message_ptr, message.byteLength);

    var res = new Uint8Array(64);
    _readBytes(signature_ptr, 64, res);

    Module._free(signature_ptr);
    Module._free(privateKey_ptr);
    Module._free(message_ptr);

    return res.buffer;
  }
  sign(privKey, message) {
    return initPromise.then(
      () =>
        new Promise(resolve => {
          resolve(this.signSync(privKey, message));
        })
    );
  }

  verifySync(pubKey, message, sig) {
    // Get a pointer to their public key
    var publicKey_ptr = _allocate(new Uint8Array(pubKey));

    // Get a pointer to the signature
    var signature_ptr = _allocate(new Uint8Array(sig));

    // Get a pointer to the message
    var message_ptr = _allocate(new Uint8Array(message));

    var res = Module._curve25519_verify(
      signature_ptr,
      publicKey_ptr,
      message_ptr,
      message.byteLength
    );

    Module._free(publicKey_ptr);
    Module._free(signature_ptr);
    Module._free(message_ptr);

    return res !== 0;
  }
  verify(pubKey, message, sig) {
    return initPromise.then(
      () =>
        new Promise((resolve, reject) => {
          if (this.verifySync(pubKey, message, sig)) {
            reject(new Error('Invalid signature'));
          } else {
            resolve();
          }
        })
    );
  }
}

export default new Curve25519Wrapper();
