import SignalProtocolAddress from '../../src/signal_protocol_address.js';
import { assertEqualArrayBuffers } from '../helpers.js';

export default function testSessionStore(store) {
  describe('SessionStore', function () {
    var number = '+5558675309';
    var testRecord = 'an opaque string';
    describe('storeSession', function () {
      var address = new SignalProtocolAddress(number, 1);
      it('stores sessions encoded as strings', function () {
        return store.storeSession(address.toString(), testRecord).then(function () {
          return store.loadSession(address.toString()).then(function (record) {
            assert.strictEqual(record, testRecord);
          });
        });
      });
      it('stores sessions encoded as array buffers', function () {
        var testRecord = new Uint8Array([1, 2, 3]).buffer;
        return store.storeSession(address.toString(), testRecord).then(function () {
          return store.loadSession(address.toString()).then(function (record) {
            assertEqualArrayBuffers(testRecord, record);
          });
        });
      });
    });
    describe('loadSession', function () {
      it('returns sessions that exist', function () {
        var address = new SignalProtocolAddress(number, 1);
        var testRecord = 'an opaque string';
        return store.storeSession(address.toString(), testRecord).then(function () {
          return store.loadSession(address.toString()).then(function (record) {
            assert.strictEqual(record, testRecord);
          });
        });
      });
      it('returns undefined for sessions that do not exist', function () {
        var address = new SignalProtocolAddress(number, 2);
        return store.loadSession(address.toString()).then(function (record) {
          assert.isUndefined(record);
        });
      });
    });
    describe('removeSession', function () {
      it('deletes sessions', function () {
        var address = new SignalProtocolAddress(number, 1);
        before(function () {
          store.storeSession(address.toString(), testRecord);
        });
        return store.removeSession(address.toString()).then(function () {
          return store.loadSession(address.toString()).then(function (record) {
            assert.isUndefined(record);
          });
        });
      });
    });
    describe('removeAllSessions', function () {
      it('removes all sessions for a number', function () {
        var devices = [1, 2, 3].map(function (deviceId) {
          var address = new SignalProtocolAddress(number, deviceId);
          return address.toString();
        });
        var promise = Promise.resolve();
        devices.forEach(function (encodedNumber) {
          promise = promise.then(function () {
            return store.storeSession(encodedNumber, testRecord + encodedNumber);
          });
        });
        return promise.then(function () {
          return store.removeAllSessions(number).then(function () {
            return Promise.all(devices.map(store.loadSession.bind(store))).then(function (records) {
              for (var i in records) {
                assert.isUndefined(records[i]);
              }
            });
          });
        });
      });
    });
  });
}
